package digi.recipeManager;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Logger;

import org.bukkit.*;
import org.bukkit.command.CommandSender;
import org.bukkit.configuration.file.*;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.inventory.*;
import org.bukkit.permissions.*;
import org.bukkit.plugin.*;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;

import digi.recipeManager.data.Item;

public class Main extends JavaPlugin
{
	protected static Server sv;
	protected static Logger log;
	protected static BukkitScheduler bs;
	protected static PluginManager pm;
	protected static Plugin plugin;
	protected static String pluginVersion;
	protected static File pluginFolder;
	
	protected static HashMap<Location, Double> furnacesSmelting = new HashMap<Location, Double>();
	
	public static boolean cfgPrintNoPermission = true;
	public static boolean cfgPrintNoWorld = true;
//	public static boolean cfgPrintFailure = true;
//	public static boolean cfgPrintClickRecipes = true;
	public static boolean cfgRemoveAllRecipes = false;
	public static int cfgFurnaceTicks = 1;
	public static String cfgLang = "en";
	
	protected static final int recipeIndicatorItemId = 2256;
	protected static final short recipeIndicatorItemData = 1337;
	
	protected static final List<String> recipeLoadErrors = new ArrayList<String>();
	private static String currentFile = null;
	private static int currentFileLine = 0;
	
	private static final HashMap<String, String> itemAliases = new HashMap<String, String>();
	
	@Override
	public void onLoad()
	{
		plugin = this;
		log = getLogger();
		sv = getServer();
		bs = sv.getScheduler();
		pm = sv.getPluginManager();
		
		final PluginDescriptionFile pluginDesc = getDescription();
		pluginVersion = pluginDesc.getVersion();
		pluginFolder = getDataFolder();
		
		loadSettings();
		
		if(!cfgRemoveAllRecipes)
		{
			// Remove all default recipes - this code is used to detect special mod
			// Mods' recipes usually start with 'X' in front or return a null recipe
			
			final Iterator<Recipe> recipes = sv.recipeIterator();
			Recipe recipe;
			ItemStack result;
			
			while(recipes.hasNext())
			{
				if((recipe = recipes.next()) == null)
					continue;
				
				if((result = recipe.getResult()) == null)
					continue;
				
				if(result.getType().toString().toLowerCase().charAt(0) != 'x')
					recipes.remove();
			}
		}
	}
	
	@Override
	public void onEnable()
	{
		getCommand("rmrecipes").setExecutor(new CmdRecipes());
		getCommand("rmcheck").setExecutor(new CmdCheck());
		getCommand("rmreload").setExecutor(new CmdReload());
		
		saveDefault("readme.txt");
		
		// Execute task after all plugins have loaded
		
		bs.scheduleSyncDelayedTask(this, new Runnable()
		{
			@Override
			public void run()
			{
				log.info("Loading settings...");
				
				loadRecipes(false);
				new Events();
			}
		});
	}
	
	@Override
	public void onDisable()
	{
		bs.cancelTasks(plugin);
		furnacesSmelting.clear();
		Recipes.reset();
		Recipes.removeCustomRecipes();
	}
	
	protected static void loadSettings()
	{
		if(pluginFolder.mkdirs())
			log.fine("Directory structure created.");
		
		saveDefault("config.yml");
		plugin.reloadConfig();
		final FileConfiguration cfg = plugin.getConfig();
		
		cfgPrintNoPermission = cfg.getBoolean("print-no-permission", cfgPrintNoPermission);
		cfgPrintNoWorld = cfg.getBoolean("print-no-world", cfgPrintNoWorld);
//		cfgPrintFailure = cfg.getBoolean("print-failed-recipe", cfgPrintFailure);
//		cfgPrintClickRecipes = cfg.getBoolean("print-click-recipes", cfgPrintClickRecipes);
		cfgRemoveAllRecipes = cfg.getBoolean("remove-all-recipes", cfgRemoveAllRecipes);
		cfgFurnaceTicks = cfg.getInt("furnace-ticks", cfgFurnaceTicks);
		cfgLang = cfg.getString("language-file", cfgLang);
		
		log.fine("config.yml settings:");
		log.fine("  print-no-permission: " + cfgPrintNoPermission);
		log.fine("  print-no-world: " + cfgPrintNoWorld);
//		log.fine("  print-failed-recipe: " + cfgPrintFailure);
//		log.fine("  print-click-recipes: " + cfgPrintClickRecipes);
		log.fine("  remove-all-recipes: " + cfgRemoveAllRecipes);
		log.fine("  furnace-ticks: " + cfgFurnaceTicks);
		log.fine("  language-file: " + cfgLang);
		
		final File aliasesFile = new File(pluginFolder + File.separator + "aliases.yml");
		
		if(!aliasesFile.exists())
			plugin.saveResource("aliases.yml", false);
		
		final FileConfiguration aliases = YamlConfiguration.loadConfiguration(aliasesFile);
		
		itemAliases.clear();
		String itemStr;
		
		for(final Entry<String, Object> entry : aliases.getValues(false).entrySet())
		{
			itemStr = entry.getValue().toString().toUpperCase();
			
			if(!itemStr.contains(":"))
				itemStr = itemStr + ":*";
			
			itemAliases.put(entry.getKey().toUpperCase(), itemStr);
		}
		
		Lang.load();
	}
	
	protected static boolean saveDefault(final String fileName)
	{
		if(!new File(pluginFolder + File.separator + fileName).exists())
		{
			plugin.saveResource(fileName, false);
			log.fine(fileName + " file created.");
			return true;
		}
		
		return false;
	}
	
	protected static boolean loadRecipes(final boolean simulation)
	{
		log.fine(simulation ? "Checking recipe files..." : "Loading recipes...");
		
		if(!simulation)
		{
			Recipes.reset();
			
			if(!cfgRemoveAllRecipes)
				Recipes.removeCustomRecipes();
			else
				sv.clearRecipes();
		}
		
		final File dir = new File(pluginFolder + File.separator + "recipes");
		
		dir.mkdirs();
		
		loadDirectory(dir, simulation);
		
		if(recipeLoadErrors.size() > 0)
		{
			final StringBuilder errors = new StringBuilder();
			
			for(final String error : recipeLoadErrors)
			{
				errors.append(error + "\r\n");
			}
			
			log.severe("There were errors processing the files: \r\n\r\n" + errors + "\r\n\r\n-- end of errors --");
			recipeLoadErrors.clear();
		}
		
		if(!simulation)
		{
			log.fine("Total loaded:");
			log.fine("  " + Recipes.craft.size() + " crafting recipes");
			log.fine("  " + Recipes.combine.size() + " combining recipes");
			log.fine("  " + Recipes.smelt.size() + " smelting recipes");
			log.fine("  " + Recipes.fuels.size() + " fuels");
			log.info("Done.");
			
			FurnaceTask.start(); // restart furnace task if any custom time smelt recipes exist
		}
		
		return (recipeLoadErrors.size() == 0);
	}
	
	private static void loadDirectory(final File dir, final boolean simulation)
	{
		log.fine("Loading '" + dir.getName() + "' folder...");
		
		String fileName;
		
		for(final File file : dir.listFiles())
		{
			fileName = file.getName();
			
			if(file.isDirectory())
			{
				if(fileName.equalsIgnoreCase("disabled"))
				{
					log.fine("Skipping the 'disabled' folder...");
					continue;
				}
				
				loadDirectory(file, simulation);
			}
			else
			{
				if(fileName.endsWith(".txt"))
				{
					try
					{
						loadDataFile(file, simulation);
					}
					catch(final Exception e)
					{
						log.severe("Failed to load '" + file.toString() + "', error:");
						e.printStackTrace();
					}
				}
			}
		}
	}
	
	private static void loadDataFile(final File file, final boolean simulation) throws IOException
	{
		final String fileName = file.getPath();
		
		log.fine("Loading '" + fileName + "' file...");
		
		final BufferedReader reader = new BufferedReader(new InputStreamReader(new DataInputStream(new FileInputStream(file))));
		final List<String> worlds = new ArrayList<String>();
		String[] split;
		String permission = null;
		String line;
		boolean header = true;
		
		while((line = reader.readLine()) != null)
		{
			currentFile = fileName;
			
			if((line = Main.processLine(line)) == null)
				continue;
			
			try
			{
				header = (header && line.charAt(0) == '@');
				
				if(header)
				{
					line = line.substring(1).trim().toLowerCase(); // remove the @ and any spaces
					
					if(line.startsWith("worlds"))
					{
						split = line.split(":");
						
						if(split.length < 2)
						{
							recipeError("@worlds doesn't have any values! (TIP: @worlds: world, world2, etc)");
							continue;
						}
						
						split = split[1].split(",");
						
						for(String world : split)
						{
							world = world.trim();
							
							if(sv.getWorld(world) == null)
							{
								recipeError("@worlds has world that doesn't exist: '" + world + "'");
								continue;
							}
							
							worlds.add(world);
						}
					}
					else if(line.startsWith("permission"))
					{
						split = line.split(":");
						
						if(split.length < 1)
						{
							recipeError("@permission doesn't have any permission node set! (TIP: @permission: permission.node)");
							continue;
						}
						
						final String[] data = split[1].split("=");
						final String perm = data[0].trim();
						
						if(perm.length() < 3 || !perm.matches("[a-zA-Z0-9.]+"))
						{
							recipeError("@permission has invalid value: '" + perm + "', minimum 3 chars, numbers, letters and dots only!");
							continue;
						}
						
						final Permission p = new Permission(perm);
						
						p.setDescription("Access to " + fileName + " recipes");
						
						if(data.length > 1)
						{
							final String defValue = data[1].trim();
							
							switch(defValue.charAt(0))
							{
								case 't':
								{
									p.setDefault(PermissionDefault.TRUE);
									break;
								}
								
								case 'f':
								{
									p.setDefault(PermissionDefault.FALSE);
									break;
								}
								
								case 'o':
								{
									p.setDefault(PermissionDefault.OP);
									break;
								}
								
								case 'n':
								{
									p.setDefault(PermissionDefault.NOT_OP);
									break;
								}
								
								default:
									recipeError("Invalid default value '" + defValue + "' for permission node '" + perm + "' (TIP: use only true/false/op/non-op)");
							}
						}
						
						permission = p.getName();
						
						try
						{
							pm.addPermission(p);
//							log.fine("Created permission node '" + permission + "' with default = " + p.getDefault().toString());
						}
						catch(final Exception e)
						{
							log.warning("Can't set default value to permission node '" + permission + "' ! (" + e.getMessage() + ")");
						}
					}
					
					continue;
				}
				
				String error = null;
				
				if(line.startsWith("CRAFT"))
					error = Recipes.craftRecipe(line, reader, permission, worlds, simulation);
				
				else if(line.startsWith("COMBINE"))
					error = Recipes.combineRecipe(line, reader, permission, worlds, simulation);
				
				else if(line.startsWith("SMELT"))
					error = Recipes.smeltRecipe(line, reader, permission, worlds, simulation);
				
				else if(line.startsWith("FUEL"))
					error = Recipes.fuelRecipe(line, reader, permission, worlds, simulation);
				
				else
					log.warning("Skipped line " + currentFileLine + ": \"" + line + "\" (TIP: forgot recipe type ? craft/combine/smelt/fuel)");
				
				if(error != null)
					recipeError("" + error);
			}
			catch(final Exception e)
			{
				e.printStackTrace();
				
				/*
				String trace = "";
				
				for(final StackTraceElement err : e.getStackTrace())
				{
					trace = trace + "\r\n    at " + err.getClassName() + "." + err.getMethodName() + "(" + err.getFileName() + ":" + err.getLineNumber() + ")";
				}
				
				recipeError("CODE EXCEPTION ERROR: " + e.getMessage() + trace);
				*/
			}
		}
		
		if(currentFileLine == 0)
			log.warning("Recipe file '" + fileName + "' is empty. (TIP: move unwanted recipe files into the 'disabled' folder)");
		
		reader.close();
	}
	
	protected static String processLine(String string)
	{
		currentFileLine++;
		
		if(string == null || string.isEmpty())
			return null;
		
		string = string.trim().toUpperCase();
		
		final String comments[] =
		{
			"//",
			"#"
		};
		
		int index;
		
		for(final String comm : comments)
		{
			index = string.indexOf(comm);
			
			if(index == 0)
				return null;
			
			if(index > -1)
				return string.substring(0, index).trim();
		}
		
		return string;
	}
	
	protected static Item processItem(String string, final int defaultData, final boolean allowAmount, final boolean allowEnchantments, final boolean printRecipeErrors)
	{
		string = string.trim();
		
		if(string.length() == 0)
		{
			if(printRecipeErrors)
				recipeError("Result string is empty!");
			
			return null;
		}
		
		final String itemString[] = string.split("\\|");
		final String stringArray[] = itemString[0].trim().split(":");
		
		Material mat = null;
		int type = 0;
		int data = defaultData;
		int amount = 1;
		
		if(stringArray.length <= 0 || stringArray[0].isEmpty())
			return new Item(0);
		
		stringArray[0] = stringArray[0].trim();
		final String alias = itemAliases.get(stringArray[0]);
		
		if(alias != null)
			return processItem(string.replace(stringArray[0], alias), defaultData, allowAmount, allowEnchantments, printRecipeErrors);
		
		mat = Material.matchMaterial(stringArray[0]);
		
		if(mat == null)
		{
			if(printRecipeErrors)
				recipeError("Invalid item name or id: '" + stringArray[0] + "', name could be different.");
			
			return null;
		}
		
		type = mat.getId();
		
		if(type <= 0) // if it's AIR we don't need to move on
			return new Item(0);
		
		if(stringArray.length > 1)
		{
			try
			{
				stringArray[1] = stringArray[1].trim();
				
				if(stringArray[1].charAt(0) != '*')
					data = Math.max(Integer.valueOf(stringArray[1]), data);
			}
			catch(final Exception e)
			{
				if(printRecipeErrors)
					recipeError("Item '" + mat + "' has invalid data value: " + stringArray[1] + ", defaulting to " + defaultData);
			}
		}
		
		if(stringArray.length > 2)
		{
			if(allowAmount)
			{
				try
				{
					amount = Math.max(Integer.parseInt(stringArray[2].trim()), 1);
				}
				catch(final Exception e)
				{
					if(printRecipeErrors)
						recipeError("Item '" + mat + "' has invalid amount value: " + stringArray[2] + ", defaulting to 1");
				}
			}
			else
			{
				if(printRecipeErrors)
					recipeError("Item '" + mat + "' can not have stack amount defined here, amount ignored");
			}
		}
		
		final Item item = new Item(type, amount, (short)data);
		
		if(itemString.length > 1)
		{
			if(allowEnchantments)
			{
				if(item.amount > 1)
				{
					if(printRecipeErrors)
						recipeError("WARNING: Item '" + mat + "' has enchantments and more than 1 item, amount set to 1.");
					
					item.amount = 1;
				}
				
				final String[] enchants = itemString[1].split(",");
				String[] enchData;
				Enchantment ench;
				int level;
				int id;
				
				for(String enchant : enchants)
				{
					enchant = enchant.trim().toUpperCase();
					enchData = enchant.split(":");
					
					if(enchData.length != 2)
					{
						if(printRecipeErrors)
							recipeError("Enchantments have to be 'ENCHANTMENT:LEVEL' format.");
						
						continue;
					}
					
					ench = Enchantment.getByName(enchData[0]);
					
					if(ench == null)
					{
						try
						{
							id = Integer.valueOf(enchData[0]);
							
							if(id <= 0)
								throw new Exception();
						}
						catch(final Exception e)
						{
							if(printRecipeErrors)
								recipeError("Invalid enchantment name or id: '" + enchData[0] + "', name could be different.");
							
							continue;
						}
						
						ench = Enchantment.getById(id);
					}
					else
						id = ench.getId();
					
					if(!ench.canEnchantItem(item.getItemStack()))
					{
						if(printRecipeErrors)
							recipeError("Can not enchantment '" + item + "' with '" + ench.toString() + "'!");
						
						continue;
					}
					
					if(enchData[1].equals("MAX"))
						level = ench.getMaxLevel();
					
					else
					{
						try
						{
							level = Integer.valueOf(enchData[1]);
							
							if(level <= 0)
								throw new Exception();
						}
						catch(final Exception e)
						{
							if(printRecipeErrors)
								recipeError("Invalid enchantment level: '" + enchData[1] + "'.");
							
							continue;
						}
						
						level = Math.min(Math.max(level, ench.getStartLevel()), ench.getMaxLevel());
					}
					
					item.enchantments.put(ench, level);
				}
			}
			else
			{
				
			}
		}
		
		return item;
	}
	
	public static void printMessage(final CommandSender to, final String[] messages)
	{
		for(String message : messages)
		{
			for(final ChatColor color : ChatColor.values())
			{
				message = message.replaceAll("(?i)<" + color.name() + ">", "" + color);
			}
		}
		
		to.sendMessage(messages);
	}
	
	public static void printMessage(final CommandSender to, String message)
	{
		for(final ChatColor color : ChatColor.values())
		{
			message = message.replaceAll("(?i)<" + color.name() + ">", "" + color);
		}
		
		to.sendMessage(message);
	}
	
	protected static void recipeError(final String error)
	{
		recipeLoadErrors.add("(" + currentFile + ":" + currentFileLine + ") " + error);
	}
}