package digi.recipeManager;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;

import org.bukkit.inventory.*;

import digi.recipeManager.data.*;
import digi.recipeManager.data.Recipe;

class Recipes
{
	protected static List<Craft> craft = new ArrayList<Craft>();
	protected static List<Combine> combine = new ArrayList<Combine>();
	protected static HashMap<String, Smelt> smelt = new HashMap<String, Smelt>();
	protected static HashMap<String, Fuel> fuels = new HashMap<String, Fuel>();
	
	protected static HashMap<String, Page> playerPage = new HashMap<String, Page>();
	
	private static int craftNum = 0;
	private static int combineNum = 0;
	protected static int smeltCustomRecipes = 0;
	
	protected static void reset()
	{
		fuels.clear();
		craft.clear();
		combine.clear();
		smelt.clear();
		
		craftNum = 0;
		combineNum = 0;
		smeltCustomRecipes = 0;
	}
	
	protected static String craftRecipe(final String line, final BufferedReader reader, final String permission, final List<String> worlds, final boolean simulation) throws Exception
	{
		String split[];
//		String itemSplit[];
		String row;
		String resultRaw = null;
		ItemData itemData;
		Character c;
		int charIndex = 97;
		int itemIndex = 0;
		int rows = 0;
		
		final HashMap<ItemData, Character> itemChars = new HashMap<ItemData, Character>();
		final List<String> recipeShape = new ArrayList<String>();
		final ItemData[] ingredients = new ItemData[9];
		final Item[] leftovers = new Item[9];
//		final StringBuilder debugMsg = new StringBuilder("CRAFT recipe\r\n    ");
		
		while(rows < 3)
		{
			// strip line of useles data, remove in-line comments and check for full-line comments
			
			if((row = Main.processLine(reader.readLine())) == null)
				continue;
			
			// if we bump into the result prematurely (smaller recipes)
			
			if(row.charAt(0) == '=')
			{
				resultRaw = row;
				break;
			}
			
			// get the items in the row, split by "+" char
			
			split = row.split("\\+");
			
			final int len = Math.min(split.length, 3); // max 3 items per row !
			final StringBuilder recipeChars = new StringBuilder();
			
			itemIndex = rows * 3;
			
			for(int i = 0; i < len; i++)
			{
				// splitting line by ">" char to get leftover item (if any)
//				itemSplit = split[i].split(">");
				
				// get main item and check/add char
				
				itemData = Main.processItem(split[i] /*itemSplit[0]*/, -1, false, false, true);
				
				if(itemData == null)
					return "One of the items is invalid, please review recipe!";
				
				if(itemData.type == 0)
					c = ' ';
				
				else
				{
					final Character ic = itemChars.get(itemData);
					
					if(ic == null)
					{
						c = (char)charIndex++;
						
						itemChars.put(itemData, c);
					}
					else
						c = ic;
					
					ingredients[itemIndex] = itemData; // add non-AIR item
				}
				
				recipeChars.append(c);
				
//				debugMsg.append((i == 0 ? "" : " + ") + (ingredients[itemIndex] == null ? "AIR" : ingredients[itemIndex].getMaterial() + ":" + ingredients[itemIndex].data));
				
				// get leftover item (if any)
				/*
				if(itemData.type != 0 && itemSplit.length > 1)
					leftovers[itemIndex] = Main.processItem(itemSplit[1].trim(), 0, true, true);
				*/
				
				itemIndex++;
			}
			
			// add to the recipe's final shape
			
			recipeShape.add(recipeChars.toString());
			
//			debugMsg.append("\r\n    ");
			
			rows++;
		}
		
		// If result wasn't found yet, it's got to be the 4th line
		
		if(resultRaw == null)
			resultRaw = Main.processLine(reader.readLine());
		
		final List<Item> results = getResults(resultRaw, reader);
		
		if(results == null)
			return "Invalid result item(s)! (TIP: forgot = character ?)";
		
/*
		for(Item result : results)
		{
			debugMsg.append("\r\n => " + result);
		}
		
		Main.log.finer(debugMsg.toString() + "\r\n");
*/		
		
		// starting the recipe and setting it's result
		
		final ShapedRecipe recipe = new ShapedRecipe(new ItemStack(Main.recipeIndicatorItemId, craftNum, Main.recipeIndicatorItemData));
		final String[] chars = new String[recipeShape.size()];
		int i = 0;
		
		for(final String shapeRow : recipeShape)
		{
			chars[i++] = shapeRow;
		}
		
		recipe.shape(chars);
		
		for(final Entry<ItemData, Character> entry : itemChars.entrySet())
		{
			recipe.setIngredient(entry.getValue(), entry.getKey().getMaterial(), entry.getKey().data);
		}
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		for(final Craft r : craft)
		{
			if(r.ingredients == ingredients)
				return "Another recipe with the same ingredients already exists!";
		}
		
		// finally add the recipe to the server!
		
		if(!Main.sv.addRecipe(recipe))
			return "Couldn't add recipe, unknown error";
		
		craft.add(craftNum, new Craft(ingredients, leftovers, results, permission, worlds));
		craftNum++;
		
		return null;
	}
	
	protected static String combineRecipe(final String line, final BufferedReader reader, final String permission, final List<String> worlds, final boolean simulation) throws Exception
	{
		final String ingrStr = Main.processLine(reader.readLine());
		
		if(ingrStr == null)
			return "Recipe has no ingredients !";
		
		final String[] ingredientsRaw = ingrStr.split("\\+");
		
		final List<Item> results = getResults(Main.processLine(reader.readLine()), reader);
		
		if(results == null)
			return "Invalid result item(s)! (TIP: forgot = character ?)";
		
		final ShapelessRecipe recipe = new ShapelessRecipe(new ItemStack(Main.recipeIndicatorItemId, combineNum, Main.recipeIndicatorItemData));
//		final StringBuilder debugMsg = new StringBuilder("COMBINE recipe\r\n    ");
		final List<Item> ingredients = new ArrayList<Item>();
		Item item;
		int items = 0;
		
		for(final String itemRaw : ingredientsRaw)
		{
			item = Main.processItem(itemRaw, -1, true, false, true);
			
			if(item == null)
				return "Combine recipes can't have AIR ingredients !";
			
//			debugMsg.append((items == 0 ? "" : " + ") + item.getMaterial() + ":" + item.data);
			
			if((items += item.amount) > 9)
				return "Combine recipes can't have more than 9 ingredients !";
			
			ingredients.add(item);
			recipe.addIngredient(item.amount, item.getMaterial(), item.data);
		}
		
/*
		for(Item result : results)
		{
			debugMsg.append("\r\n => " + result);
		}
		
		Main.log.finer(debugMsg.toString() + "\r\n");
*/		
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		for(final Combine r : combine)
		{
			if(r.ingredients == ingredients)
				return "Another recipe with the same ingredients already exists!";
		}
		
		if(!Main.sv.addRecipe(recipe))
			return "Couldn't add recipe, unknown error";
		
		combine.add(combineNum, new Combine(ingredients, results, permission, worlds));
		combineNum++;
		
		return null;
	}
	
	protected static String smeltRecipe(final String line, final BufferedReader reader, final String permission, final List<String> worlds, final boolean simulation) throws Exception
	{
		final String[] split = Main.processLine(reader.readLine()).split("%");
		
		if(split.length == 0)
			return "No ingredient!";
		
		final Item result = getResult(Main.processLine(reader.readLine()), reader);
		
		if(result == null)
			return "Invalid result item! (TIP: forgot = character ?)";
		
		final Item ingredient = Main.processItem(split[0], -1, false, false, true);
		double minTime = -1.0;
		double maxTime = -1.0;
		
		if(split.length >= 2)
		{
			final String[] timeSplit = split[1].trim().split("-");
			
			if(!timeSplit[0].equals("INSTANT"))
				minTime = Double.valueOf(timeSplit[0]);
			
			if(timeSplit.length >= 2)
				maxTime = Double.valueOf(timeSplit[1]);
			
			if(maxTime > -1.0 && minTime >= maxTime)
				return "Smelting recipe has minimum time less or equal to maximum time! (TIP: Use a single number if you want a fixed value)";
		}
		
		// the recipe
		
//		Main.log.finer("SMELT: " + ingredient.getMaterial() + ":" + ingredient.data + " burns for " + (maxTime > minTime ? minTime + "-" + maxTime : minTime) + "s and results " + result.toString() + " " + (permission != null ? " with permission = " + permission : ""));
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		if(Recipes.smelt.containsKey(ingredient.type + (ingredient.data == -1 ? "" : ":" + ingredient.data)))
			return "Another smelting recipe with this ingredient already exists!";
		
		if(!Main.sv.addRecipe(new FurnaceRecipe((result.type == 0 ? new ItemStack(3) : result.getItemStack()), ingredient.getMaterial(), ingredient.data)))
			return "Couldn't add recipe, unknown error";
		
		smelt.put(getIngredientString(ingredient), new Smelt(ingredient, result, minTime, maxTime, permission, worlds));
		
		if(minTime >= 0.0)
			smeltCustomRecipes++;
		
		return null;
	}
	
	protected static String fuelRecipe(final String line, final BufferedReader reader, final String permission, final List<String> worlds, final boolean simulation) throws Exception
	{
		final String[] split = Main.processLine(reader.readLine()).split("%");
		
		if(split[1] == null)
			return "No burn time set!";
		
		final String[] timeSplit = split[1].trim().split("-");
		final int minTime = Math.max(Integer.parseInt(timeSplit[0]), 1);
		int maxTime = -1;
		
		if(timeSplit.length >= 2)
			maxTime = Math.max(Integer.parseInt(timeSplit[1]), maxTime);
		
		final ItemData ingredient = Main.processItem(split[0], -1, false, false, true);
		
		if(ingredient == null || ingredient.type == 0)
			return "Invalid item: '" + ingredient + "'";
		
		if(minTime <= 0)
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " can't burn for negative or zero seconds!";
		
		if(maxTime > -1 && minTime >= maxTime)
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " has minimum time less or equal to maximum time! (TIP: Use a single number if you want a fixed value)";
		
//		Main.log.finer("FUEL: " + ingredient.getMaterial() + ":" + ingredient.data + " burns for " + (maxTime > minTime ? minTime + "-" + maxTime : minTime) + "s " + (permission != null ? " with permission = " + permission : ""));
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		if(fuels.containsKey(ingredient))
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " already exists!";
		
		fuels.put(getIngredientString(ingredient), new Fuel(ingredient, minTime, maxTime, permission, worlds));
		
		return null;
	}
	
	private static Item getResult(final String line, final BufferedReader reader) throws Exception
	{
		final List<Item> results = getResults(line, reader, true, true);
		
		if(results == null || results.size() < 1)
			return null;
		else if(results.size() > 1)
			Main.log.warning("Can't have more than 1 result, the rest were ignored.");
		
		return results.get(0);
	}
	
	private static List<Item> getResults(final String line, final BufferedReader reader) throws Exception
	{
		return getResults(line, reader, false, false);
	}
	
	private static List<Item> getResults(String line, final BufferedReader reader, final boolean allowAir, final boolean oneResult) throws Exception
	{
		final List<Item> results = new ArrayList<Item>();
		Item noPercentItem = null;
		String[] resultRawSplit;
		Item item;
		int totalpercentage = 0;
		
		while(line != null && !line.isEmpty() && line.charAt(0) == '=')
		{
			resultRawSplit = line.substring(1).trim().split("%"); // remove the = character and see if it's got percentage
			
			if(resultRawSplit.length >= 2)
			{
				item = Main.processItem(resultRawSplit[1].trim(), 0, true, true, true);
				
				if(item == null || (!allowAir && item.type == 0))
				{
					Main.recipeError("Invalid result !");
					return null;
				}
				
				try
				{
					item.chance = Math.min(Math.max(Integer.valueOf(resultRawSplit[0].trim()), 1), 100);
				}
				catch(final Exception e)
				{
					Main.recipeError("Invalid percentage number: " + resultRawSplit[0]);
					return null;
				}
				
				if((totalpercentage += item.chance) > 100)
				{
					Main.recipeError("Total result items' chance exceeds 100% ! (TIP: not defining percentage for 1 item will make it's chance fit with the rest until 100%)");
					return null;
				}
				
				results.add(item);
			}
			else
			{
				if(resultRawSplit[0] == null)
					return null;
				
				if(noPercentItem != null)
				{
					Main.recipeError("Can't have more than 1 item without procentage to fill the rest!");
					return null;
				}
				
				noPercentItem = Main.processItem(resultRawSplit[0].trim(), 0, true, true, true);
				
				if(noPercentItem == null || (!allowAir && noPercentItem.type == 0))
				{
					Main.recipeError("Invalid result !");
					return null;
				}
			}
			
			line = Main.processLine(reader.readLine());
		}
		
		if(noPercentItem != null)
		{
			noPercentItem.chance = (100 - totalpercentage);
			results.add(noPercentItem);
		}
		else if(!oneResult && totalpercentage < 100)
		{
			results.add(new Item(0, 0, (short)0, (100 - totalpercentage)));
		}
		
		return results;
	}
	
	private static String getIngredientString(final ItemData ingredient)
	{
		return ingredient.type + (ingredient.data == -1 ? "" : ":" + ingredient.data);
	}
	
	protected static Smelt getSmeltRecipe(final ItemStack ingredient)
	{
		final ItemData item = new ItemData(ingredient);
		final Smelt recipe = Recipes.smelt.get(item.type + "");
		
		if(recipe == null)
			return Recipes.smelt.get(item.type + ":" + item.data);
		
		return recipe;
	}
	
	protected static Fuel getFuelRecipe(final ItemStack ingredient)
	{
		final ItemData item = new ItemData(ingredient);
		final Fuel recipe = Recipes.fuels.get(item.type + "");
		
		if(recipe == null)
			return Recipes.fuels.get(item.type + ":" + item.data);
		
		return recipe;
	}
	
	protected static void removeCustomRecipes()
	{
		final Iterator<org.bukkit.inventory.Recipe> recipes = Main.sv.recipeIterator();
		org.bukkit.inventory.Recipe recipe;
		
		while(recipes.hasNext())
		{
			if((recipe = recipes.next()) == null)
				continue;
			
			if(Recipes.isCustomRecipe(recipe.getResult()))
				recipes.remove();
		}
	}
	
	protected static boolean isCustomRecipe(final ItemStack result)
	{
		return (result != null && result.getTypeId() == Main.recipeIndicatorItemId && result.getDurability() == Main.recipeIndicatorItemData);
	}
	
	protected static List<Recipe> getRecipesForItem(final Item item, final boolean ingredient)
	{
		final List<Recipe> recipes = new ArrayList<Recipe>();
		
		for(final Craft r : craft)
		{
			if(ingredient)
			{
				for(final ItemData i : r.ingredients)
				{
					if(item.compareItemData(i))
					{
						recipes.add(r);
						break;
					}
				}
			}
			else
			{
				for(final Item i : r.results)
				{
					if(item.compareItemData(i))
					{
						recipes.add(r);
						break;
					}
				}
			}
		}
		
		for(final Combine r : combine)
		{
			for(final ItemData i : (ingredient ? r.ingredients : r.results))
			{
				if(item.compareItemData(i))
				{
					recipes.add(r);
					break;
				}
			}
		}
		
		for(final Smelt r : smelt.values())
		{
			if(item.compareItemData(ingredient ? r.ingredient : r.result))
				recipes.add(r);
		}
		
		if(ingredient)
		{
			for(final Fuel r : fuels.values())
			{
				if(item.compareItemData(r.ingredient))
					recipes.add(r);
			}
		}
		
		return recipes;
	}
}